/* % Nonlinear simulator for spacecraft rendezvous 
 *   based on the simulator described by Mounir 
 *   KARA-ZAITRI dans "Modélisation et guidage robuste
 *   et autonome pour le problème du rendez-vous orbital 
 * 
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 *  Author      : Paulo Ricardo ARANTES GILZ
 *  Tuteurs     : Mioara JOLDES / Christophe LOUEMBET
 *  Date        : 2016/08/01
 *  Copyright   : LAAS-CNRS - Toulouse/FRANCE
 * %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
 */

///////////////////////////////////////////
// MACROS
///////////////////////////////////////////

# define  dbl               double
# define  pow2(x)    		(((x)*(x)))

///////////////////////////////////////////
// GLOBAL VARIABLES
///////////////////////////////////////////

dbl t0, tf, step, out_freq;
dbl mu, Re, J2, m_l, m_f, ds_l, ds_f, dc_l, dc_f;
dbl a_l, e_l, i_l, Omega_l, omega_l, nu_l;
dbl a_f, e_f, i_f, Omega_f, omega_f, nu_f;
dbl atm_drag, J2_effect, solar_actvty;

struct timeval t_init;
struct timeval t_final;

///////////////////////////////////////////
// FUNCTION HEADER
///////////////////////////////////////////

void    		tic(void);
void 			toc(void);
void 			read_inputs(void);
void 			eq2in(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * xin, dbl * yin, dbl * zin, dbl * dxin, dbl * dyin, dbl * dzin );
void 			c2eq(dbl a, dbl e, dbl i, dbl Omega, dbl omega, dbl nu, dbl *p, dbl *f, dbl *g, dbl *h, dbl *k, dbl *L);
void 			accJ2(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * ar, dbl * as, dbl * aw);
void 			accAD(int lorf, dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * ar, dbl * as);
dbl 			rho(dbl R);
void 			deq(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl ar, dbl as, dbl az, dbl *dp, dbl *df, dbl *dg, dbl *dh, dbl *dk, dbl *dL);
void 			integrate(void);
void 			lvlh_rel(dbl p_l, dbl f_l, dbl g_l, dbl h_l, dbl k_l, dbl L_l, dbl p_f, dbl f_f, dbl g_f, dbl h_f, dbl k_f, dbl L_f, dbl * x_lvlh, dbl * y_lvlh, dbl * z_lvlh, dbl * dx_lvlh, dbl * dy_lvlh, dbl * dz_lvlh);

///////////////////////////////////////////
// FUNCTION BODY
///////////////////////////////////////////

// Initializing clock 
void    tic(void){
	gettimeofday(&t_init, 0);
}

// Stopping clock and printing elapsed time
void    toc(void){
	gettimeofday(&t_final,0);
	long long elapsed = (t_final.tv_sec-t_init.tv_sec)*1000000LL + t_final.tv_usec-t_init.tv_usec;

	printf("Elapsed time: %lld microseconds \n",elapsed);
}

// Reading inputs
void read_inputs(void){
	
	int i = 0;		
	dbl inputs[28] = {0};

	FILE * fp;
	fp = fopen("inputs","r");

	if (fp == NULL){
	    printf("Error opening file to be read.\n");
	    exit(-1);
	}

	char buff[500];	
	while ( fgets(buff,500,fp)!=NULL ){		
		sscanf(buff,"%lf",&inputs[i]);			
		i++;
	}
	fclose(fp);

	t0			= inputs[0];				// Initial time
	tf 			= inputs[1];				// Final time
	step		= inputs[2];				// Integration step

	out_freq    = inputs[3];                // Outputting frequency 

 	mu 			= inputs[4];    			// Earths gravitational constant
 	Re 			= inputs[5];     			// Earths radius
 	J2  		= inputs[6];     			// J2 effect coefficient

 	m_l  		= inputs[7];     			// Mass of the leader spacecraft 
 	m_f			= inputs[8];     			// Mass of the follower spacecraft
 	ds_l 		= inputs[9];     			// Drag surface of the leader spacecraft
 	ds_f 		= inputs[10];     			// Drag surface of the follower spacecraft
 	dc_l 		= inputs[11];     			// Drag coefficient of the leader spacecraft
 	dc_f 		= inputs[12];     			// Drag coefficient of the follower spacecraft

 	a_l			=inputs[13]; 				// Semi-major axis (a) leader
 	e_l			=inputs[14];				// Excentricity (e) leader
 	i_l			=inputs[15];      			// Inclination (i) leader
 	Omega_l 	=inputs[16];      			// Longitude of ascending node (Omega) leader
 	omega_l 	=inputs[17];      			// Argument of periapsis (omega) leader 	 
 	nu_l 		=inputs[18];      			// True anomaly (nu) leader

 	a_f			=inputs[19]; 				// Semi-major axis (a) follower
 	e_f			=inputs[20];				// Excentricity (e) follower
 	i_f			=inputs[21];      			// Inclination (i) follower
 	Omega_f 	=inputs[22];      			// Longitude of ascending node (Omega) follower
 	omega_f 	=inputs[23];      			// Argument of periapsis (omega) follower 	 
 	nu_f 		=inputs[24];      			// True anomaly (nu) follower

 	J2_effect	 = inputs[25];  			// J2 effect (1 - yes, 0 - no)
 	atm_drag     = inputs[26];  			// Atmospheric drag effect (1 - yes, 0 - no) 	
 	solar_actvty = inputs[27];  			// Solar activity (1 - hi, 0 - lo)

	int err = 0;

 	if (t0 < 0){
 		printf("Error: initial time is strictly negative. \n"); 		
 		err = 1;
 	}

	if ( (tf < 0) || (tf <= t0) ){
 		printf("Error: final time is negative or lower than or equal to initial time. \n"); 		
 		err = 1;
 	}

	if ( (step <= 0) || (step > tf-t0) ){
 		printf("Error: step must be strictly positive and greter than (tf-t0). \n"); 		
 		err = 1;
 	}


 	if ( (round(out_freq) != out_freq) || (out_freq < 1) ){
 		printf("Error: outputting frequency must be integer greater or equal than 1. \n"); 		
 		err = 1;
 	}


 	if ( mu<=0 ){
 		printf("Error: earths gravitational is negative or null. \n"); 		
 		err = 1;
 	}

 	if ( Re<=0 ){
 		printf("Error: earths radius is negative or null. \n");
 		err = 1; 		
 	}

 	if ( m_l<=0 ){
 		printf("Error: mass of leader spacecraft is negative or null. \n");
 		err = 1; 		
 	}

 	if ( m_f<=0 ){
 		printf("Error: mass of follower spacecraft is negative or null. \n");
 		err = 1; 		
 	}

 	if ( ds_l<=0 ){
 		printf("Error: drag surface of leader spacecraft is negative or null. \n");
 		err = 1; 		
 	}

 	if ( ds_f<=0 ){
 		printf("Error: drag surface of follower spacecraft is negative or null. \n");
 		err = 1; 		
 	}

 	if ( dc_l<=0 ){
 		printf("Error: drag coefficient of leader spacecraft is negative or null. \n");
 		err = 1; 		
 	}

 	if ( dc_f<=0 ){
 		printf("Error: drag coefficient of follower spacecraft is negative or null. \n");
 		exit(-1);
 	}

 	if (a_l<=0){
 		printf("Error: semi-major axis of leader spacecraft is negative or null. \n");
 		exit(-1);
 	}

 	if ( a_f<=0 ){
 		printf("Error: semi-major axis of follower spacecraft is negative or null. \n");
 		exit(-1);
 	}

	if ( (e_l<=0) || (e_l>=1) ){
 		printf("Error: excentricity axis of leader spacecraft is negative or null. \n"); 		
 	}

	if ( (e_f<=0) || (e_f>=1) ){
 		printf("Error: excentricity axis of follower spacecraft is negative or null. \n"); 		
 	}

 	if ( !((atm_drag == 1) || (atm_drag==0)) ){
 		printf("Error: atmospheric drag must be 0 or 1. \n"); 		
 	}

	if ( !((J2_effect == 1) || (J2_effect==0)) ){
 		printf("Error: J2 effect must be 0 or 1. \n"); 		
 	}

 	if ( !((solar_actvty == 1) || (solar_actvty==0)) ){
 		printf("Error: solar activity must be 0 or 1. \n"); 		
 	}

 	if (err==1){
 		exit(-1);
 	}
}

// Changing from equinotial to inertial
void eq2in(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * xin, dbl * yin, dbl * zin, dbl * dxin, dbl * dyin, dbl * dzin ){

	dbl w 	   = 1+f*cos(L)+g*sin(L);
	dbl s 	   = sqrt(1+h*h+k*k);
	dbl alpha2 = h*h-k*k;
	dbl r 	   = p/w;

    *xin = r/(s*s)*(cos(L)+alpha2*cos(L)+2*h*k*sin(L));
    *yin = r/(s*s)*(sin(L)-alpha2*sin(L)+2*h*k*cos(L));
    *zin = 2*r/(s*s)*(h*sin(L)-k*cos(L));
    *dxin = -1/(s*s)*sqrt(mu/p)*(sin(L)+alpha2*sin(L)-2*h*k*cos(L)+g-2*f*h*k+alpha2*g);
    *dyin = -1/(s*s)*sqrt(mu/p)*(-cos(L)+alpha2*cos(L)+2*h*k*sin(L)-f+2*g*h*k+alpha2*f);
    *dzin = 2/(s*s)*sqrt(mu/p)*(h*cos(L)+k*sin(L)+f*h+g*k);
}

// Changing from classic to equinotial
void c2eq(dbl a, dbl e, dbl i, dbl Omega, dbl omega, dbl nu, dbl *p, dbl *f, dbl *g, dbl *h, dbl *k, dbl *L){
	
	*p = a*(1-e*e);
	*f = e*cos(Omega+omega);
	*g = e*sin(Omega+omega);
	*h = tan(i/2)*cos(Omega);
	*k = tan(i/2)*sin(Omega);
	*L = Omega + omega + nu;
}

// J2 acceleration
void accJ2(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * ar, dbl * as, dbl * aw){

	dbl w     = 1+f*cos(L)+g*sin(L);
	dbl r     = p/w;
	dbl coeff = -mu*J2*Re*Re/(r*r*r*r);

	if (J2_effect==1){
		*ar = coeff * 3/2 * ( 1 - 12*pow2(h*sin(L)-k*cos(L))/pow2(1+h*h+k*k)  );
		*as = coeff * 12  * ((h*sin(L)-k*cos(L))*(h*cos(L)+k*sin(L))/pow2(1+h*h+k*k) );	
		*aw = coeff *  6  * ((h*sin(L)-k*cos(L))*(1-h*h-k*k)/pow2(1+h*h+k*k) );	
	}
	else{		
		*ar = 0;
		*as = 0;	
		*aw = 0;		
	}	
}

// Atmospheric drag acceleration
void accAD(int lorf, dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl * ar, dbl * as){

	// lorf: 1 if leader, 0 if follower

	dbl xin, yin, zin, dxin, dyin, dzin;	
	dbl m, C, S;
	dbl R, V, coeff;

	if (atm_drag == 1){
		eq2in(p, f, g, h, k, L, &xin, &yin, &zin, &dxin, &dyin, &dzin);

		if(lorf==1){
			m = m_l;
			C = dc_l;
			S = ds_l;
		}
		else{
			m = m_f;
			C = dc_f;
			S = ds_f;
		}
		
		R = sqrt(xin*xin + yin*yin + zin*zin);
		V = sqrt(dxin*dxin + dyin*dyin + dzin*dzin);
		coeff = -0.5 * rho(R) * S * C / m;

		*ar = coeff * sqrt(mu/p) * V * (f*sin(L)-g*cos(L));
		*as = coeff * sqrt(mu/p) * V * (1 + f*cos(L)+g*sin(L));
	}
	else{		
		*ar = 0;
		*as = 0;	
	}
}

// Air density model
dbl rho(dbl R){

	dbl rholh, val;	

	if (solar_actvty == 1){
		rholh = 3.5475e-11;
	}
	else{
		rholh = 2.2644e-12;
	}

	if (R>4e5+Re){
		val = (rholh * exp( -1*(R-4e5-Re)/46830 ));
	}
	else{
		val = rholh;
	}

	return val;
}

// Derivatives of equinotial orbital elements
void deq(dbl p, dbl f, dbl g, dbl h, dbl k, dbl L, dbl ar, dbl as, dbl aw, dbl *dp, dbl *df, dbl *dg, dbl *dh, dbl *dk, dbl *dL){

	dbl w = 1+f*cos(L)+g*sin(L);
	dbl s = sqrt(1+h*h+k*k);

	*dp =                           (2*p/(w)*sqrt(p/mu))*as;
	*df = (sqrt(p/mu)*sin(L))*ar  + (sqrt(p/mu)*1/w*((1+w)*cos(L)+f))*as + (-sqrt(p/mu)*g/w*(h*sin(L)-k*cos(L)))*aw;
    *dg = (-sqrt(p/mu)*cos(L))*ar + (sqrt(p/mu)*1/w*((1+w)*sin(L)+g))*as + (sqrt(p/mu)*f/w*(h*sin(L)-k*cos(L)))*aw;
    *dh =                                                                  (sqrt(p/mu)*s*s*cos(L)/(2*w))*aw;
   	*dk =                                                                  (sqrt(p/mu)*s*s*sin(L)/(2*w))*aw;
   	*dL =                                                                  ((1/w)*sqrt(p/mu)*(h*sin(L)-k*cos(L)))*aw + sqrt(mu*p)*(w*w)/(p*p);
}

// Integration loop
void integrate(void){	

	int out;

	dbl p_l, f_l, g_l, h_l, k_l, L_l;
	dbl p_f, f_f, g_f, h_f, k_f, L_f;	
	dbl x_lvlh, y_lvlh, z_lvlh, dx_lvlh, dy_lvlh, dz_lvlh;	

	c2eq(a_l, e_l, i_l, Omega_l, omega_l, nu_l, &p_l, &f_l, &g_l, &h_l, &k_l, &L_l);
	c2eq(a_f, e_f, i_f, Omega_f, omega_f, nu_f, &p_f, &f_f, &g_f, &h_f, &k_f, &L_f);

	dbl tc = t0;

	dbl arJ2_l, asJ2_l, awJ2_l, arAD_l, asAD_l;
	dbl arJ2_f, asJ2_f, awJ2_f, arAD_f, asAD_f; 
	dbl dp_l, df_l, dg_l, dh_l, dk_l, dL_l;
	dbl dp_f, df_f, dg_f, dh_f, dk_f, dL_f;	

	FILE *file = fopen("outputs", "w");
	if (file == NULL){
	    printf("Error creating output file.\n");
	    exit(-1);
	}	

	out = 1;
	while(tc<tf){
	
		accJ2(p_l, f_l, g_l, h_l, k_l, L_l, &arJ2_l, &asJ2_l, &awJ2_l);
		accAD(1, p_l, f_l, g_l, h_l, k_l, L_l, &arAD_l, &asAD_l);
		deq(p_l, f_l, g_l, h_l, k_l, L_l, (arJ2_l+arAD_l), (asJ2_l+asAD_l), awJ2_l, &dp_l, &df_l, &dg_l, &dh_l, &dk_l, &dL_l);
		p_l += step* dp_l;
		f_l += step* df_l;
		g_l += step* dg_l;
		h_l += step* dh_l;
		k_l += step* dk_l;
		L_l += step* dL_l;

		accJ2(p_f, f_f, g_f, h_f, k_f, L_f, &arJ2_f, &asJ2_f, &awJ2_f);
		accAD(0, p_f, f_f, g_f, h_f, k_f, L_f, &arAD_f, &asAD_f);
		deq(p_f, f_f, g_f, h_f, k_f, L_f, (arJ2_f+arAD_f), (asJ2_f+asAD_f), awJ2_f, &dp_f, &df_f, &dg_f, &dh_f, &dk_f, &dL_f);
		p_f += step* dp_f;
		f_f += step* df_f;
		g_f += step* dg_f;
		h_f += step* dh_f;
		k_f += step* dk_f;
		L_f += step* dL_f;

		lvlh_rel(p_l, f_l, g_l, h_l, k_l, L_l, p_f, f_f, g_f, h_f, k_f, L_f, &x_lvlh, &y_lvlh, &z_lvlh, &dx_lvlh, &dy_lvlh, &dz_lvlh);				

		// Step increment
		tc += step;

		// Printing result to output file within the desired frequency
		if((out==out_freq) || (tc == t0+step) || (tc>=tf)){
			fprintf(file, "%.53f %.53f %.53f %.53f %.53f %.53f \n", x_lvlh, y_lvlh, z_lvlh, dx_lvlh, dy_lvlh, dz_lvlh);	
		}
		
		if (out+1>out_freq){
			out=1;			
		}
		else{
			out+=1;
		}							
	}

	fclose(file);
}

// Computing lvlh relative state from equinotial parameters and printing to file  
void lvlh_rel(dbl p_l, dbl f_l, dbl g_l, dbl h_l, dbl k_l, dbl L_l, dbl p_f, dbl f_f, dbl g_f, dbl h_f, dbl k_f, dbl L_f, dbl * x_lvlh, dbl * y_lvlh, dbl * z_lvlh, dbl * dx_lvlh, dbl * dy_lvlh, dbl * dz_lvlh){

	dbl xin_l, yin_l, zin_l, dxin_l, dyin_l, dzin_l;	
	dbl xin_f, yin_f, zin_f, dxin_f, dyin_f, dzin_f;
	dbl xin_rel, yin_rel, zin_rel, dxin_rel, dyin_rel, dzin_rel;
	dbl urx, ury, urz;
	dbl usx, usy, usz;
	dbl utx, uty, utz;
	dbl uwx, uwy, uwz;
	dbl r, s, w, dr, ds, dw;
	dbl vrx, vry;
	dbl norm;

	eq2in(p_l, f_l, g_l, h_l, k_l, L_l, &xin_l, &yin_l, &zin_l, &dxin_l, &dyin_l, &dzin_l);	
	eq2in(p_f, f_f, g_f, h_f, k_f, L_f, &xin_f, &yin_f, &zin_f, &dxin_f, &dyin_f, &dzin_f);

	xin_rel = xin_f - xin_l;
	yin_rel = yin_f - yin_l;
	zin_rel = zin_f - zin_l;
	dxin_rel = dxin_f - dxin_l;
	dyin_rel = dyin_f - dyin_l;
	dzin_rel = dzin_f - dzin_l;

	urx = (xin_l)/sqrt(pow2(xin_l) + pow2(yin_l) + pow2(zin_l)); 
	ury = (yin_l)/sqrt(pow2(xin_l) + pow2(yin_l) + pow2(zin_l));
	urz = (zin_l)/sqrt(pow2(xin_l) + pow2(yin_l) + pow2(zin_l));

	utx = (dxin_l)/sqrt(pow2(dxin_l) + pow2(dyin_l) + pow2(dzin_l)); 
	uty = (dyin_l)/sqrt(pow2(dxin_l) + pow2(dyin_l) + pow2(dzin_l));
	utz = (dzin_l)/sqrt(pow2(dxin_l) + pow2(dyin_l) + pow2(dzin_l));

	uwx = - uty*urz + utz*ury;
	uwy = - utz*urx + utx*urz;
	uwz = - utx*ury + uty*urx;

	usx =  uwy*urz - uwz*ury;
	usy =  uwz*urx - uwx*urz;
	usz =  uwx*ury - uwy*urx;	

	norm = sqrt(pow2(uwx) + pow2(uwy) + pow2(uwz));
	uwx /= norm;
	uwy /= norm;
	uwz /= norm;

	norm = sqrt(pow2(usx) + pow2(usy) + pow2(usz));
	usx /= norm;
	usy /= norm;
	usz /= norm;

	r = xin_rel*urx + yin_rel*ury + zin_rel*urz;
	s = xin_rel*usx + yin_rel*usy + zin_rel*usz;
	w = xin_rel*uwx + yin_rel*uwy + zin_rel*uwz;

	*x_lvlh =  s;
	*y_lvlh =  w; 
	*z_lvlh = -r;

	dbl nudot = sqrt(mu*p_l)*pow2( (1+f_l*cos(L_l)+g_l*sin(L_l))/p_l );

	vrx = -nudot*s;
	vry =  nudot*r;

	dr = dxin_rel*urx + dyin_rel*ury + dzin_rel*urz - vrx;
	ds = dxin_rel*usx + dyin_rel*usy + dzin_rel*usz - vry;
	dw = dxin_rel*uwx + dyin_rel*uwy + dzin_rel*uwz;

	*dx_lvlh =  ds;
	*dy_lvlh =  dw; 
	*dz_lvlh = -dr;	
}